import logging
from concurrent.futures import Future
from dataclasses import dataclass, field
from functools import partial
from threading import Thread
from typing import Dict
from uuid import uuid4

import dill
from pika import BasicProperties, BlockingConnection
from pika.adapters.blocking_connection import BlockingChannel
from retry import retry

logging.getLogger("pika").setLevel(logging.WARN)
logger = logging.getLogger(__name__)


@dataclass
class Client:
    server_queue: str
    conn: BlockingConnection
    channel: BlockingChannel
    timeout: int
    _thread: Thread
    _waiting: Dict[str, Future] = field(default_factory=dict)

    def __post_init__(self):
        self.channel.basic_consume("amq.rabbitmq.reply-to", self.recieve, auto_ack=True)

    def call(self, method, *args, **kwargs):
        f: Future = Future()
        key = uuid4().hex
        self._waiting[key] = f
        self.conn.add_callback_threadsafe(
            lambda: self.channel.basic_publish(
                exchange="",
                routing_key=self.server_queue,
                body=dill.dumps(
                    {"method": method, "args": args, "kwargs": kwargs, "key": key}
                ),
                properties=BasicProperties(reply_to="amq.rabbitmq.reply-to"),
            )
        )
        return f.result(timeout=self.timeout)

    def __getattr__(self, method):
        return partial(self.call, method)

    def recieve(self, ch, method_frame, properties, body):
        body = dill.loads(body)
        f = self._waiting.pop(body["key"])
        if "body" in body:
            f.set_result(body["body"])
        elif "exception" in body:
            f.set_exception(body["exception"])
        else:
            raise RuntimeError()


def get_client(server_queue: str, server_connection_parameters, timeout=10) -> Client:
    conn = BlockingConnection(server_connection_parameters)
    channel = conn.channel()

    @retry()
    def worker():
        logger.debug('starting worker listening on %s', server_queue)
        try:
            channel.start_consuming()
        except Exception as e:
            logger.exception('worker died')

    t = Thread(target=worker)
    t.daemon = True
    t.start()

    return Client(server_queue, conn, channel, timeout, _thread=t)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mause_rpc']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.1,<0.4.0', 'pika>=1.1.0,<2.0.0', 'retry>=0.9.2,<0.10.0']

setup_kwargs = {
    'name': 'mause-rpc',
    'version': '0.0.13',
    'description': 'A dumb as hell rpc implementation built on rabbitmq',
    'long_description': "Mause RPC\n=========\n\nA dumb as hell rpc implementation built on rabbitmq\n\nNeed to write a server?\n\n```py\nfrom mause_rpc.server import Server\n\nrpc_queue = 'rpc.queue'\nserver = Server(rpc_queue, 'rabbitmq://...')\n\n\n@server.register\ndef hello(name: str) -> str:\n    return 'hello ' + name\n\n\n@server.register\ndef div(a: int, b: int) -> float:\n    if b == 0:\n        raise ZeroDivisionError()\n    return a / b\n\n\nif __name__ == '__main__':\n    server.serve()\n\n```\n\nNeed a client?\n\n```py\nfrom mause_rpc.client import get_client\n\nrpc_queue = 'rpc.queue'\nclient = get_client(rpc_queue, 'rabbitmq://...')\n\n\ndef test_basic_functionality():\n    assert client.hello('mark') == 'hello mark'\n    assert client.div(5, 2) == 2.5\n\n    with pytest.raises(ZeroDivisionError):\n        client.div(5, 0)\n```\n",
    'author': 'Elliana May',
    'author_email': 'me@mause.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Mause/rpc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import socket
from dataclasses import dataclass, field
from functools import partial
from typing import Callable, Dict, Union

import dill
import pika
from pika import ConnectionParameters
from pika.exceptions import AMQPConnectionError, ChannelClosedByBroker
from retry import retry

logging.basicConfig(level=logging.INFO)
logging.getLogger("pika").setLevel(logging.WARN)


@dataclass
class Server:
    server_queue: str
    connection_params: ConnectionParameters
    _methods: Dict[str, Callable] = field(default_factory=dict)

    def register(self, method: Union[str, Callable]):
        if isinstance(method, str):
            return partial(self._register, method)
        else:
            return self._register(method.__name__, method)

    def _register(self, name, method):
        self._methods[name] = method
        return method

    @retry(socket.gaierror, delay=10, jitter=(1, 3))
    @retry(ChannelClosedByBroker, delay=10, jitter=(1, 3))
    @retry(AMQPConnectionError, delay=5, jitter=(1, 3))
    def serve(self):
        with pika.BlockingConnection(self.connection_params) as conn:
            channel = conn.channel()

            channel.queue_declare(
                queue=self.server_queue, exclusive=True, auto_delete=True
            )
            channel.basic_consume(self.server_queue, self.on_server_rx_rpc_request)
            logging.info("Ready, waiting on work on %s", self.server_queue)
            channel.start_consuming()

    def on_server_rx_rpc_request(self, ch, method_frame, properties, body):
        body = dill.loads(body)
        logging.info("RPC Server got request: %s", body)

        res = {"key": body["key"]}

        try:
            res["body"] = self._methods[body["method"]](*body["args"], **body["kwargs"])
        except Exception as e:
            logging.exception("Call to %s caused exception", body["method"])
            res["exception"] = e

        ch.basic_publish("", routing_key=properties.reply_to, body=dill.dumps(res))

        ch.basic_ack(delivery_tag=method_frame.delivery_tag)
